//
// QSINIT EFI loader
// main() function
//
#include "qsloader.h"
#include "efilib.h"
#include "pe.h"
#include "seldesc.h"

#define RESERVE_DESCRIPTORS    256
#define PRINT_MEM_TABLE        0
#define XCPT_STACK_SIZE        (48*1024)

#pragma pack(1)
typedef struct {
   UINT16                limit;
   EFI_PHYSICAL_ADDRESS   base;
} GDT64;
#pragma pack()

void ReadGDTR       (GDT64 *Gdt);
void WriteGDTR      (GDT64 *Gdt);
void ReadIDTR       (GDT64 *Gdt);
void SaveSegRegs    (void);
void RestSegRegs    (void);

EFI_HANDLE              SelfImg = 0;
static GDT64           SavedGDT,
                         CurGDT,
                         CurIDT;
u64t                   SavedCr4;
u32t*                     lapic = 0;       // non-zero only in MT mode
u32t                   SavedSpr = 0;
EFI_PHYSICAL_ADDRESS    GDTCopy = 0;       // own GDT :)
UINT32                  GDTSize = 0,       // size of GDT in bytes
                      GDTLowest = 0,       // lowest usable selector
                   GDTCopyPages = 0;       // # of AllocPages to free
EFI_PHYSICAL_ADDRESS      QSMem = 0;       // memory for QSINIT
UINTN                   QSPages = 0;       // size of it in pages
EFI_LOADED_IMAGE   *LoadedImage = 0;
EFI_DEVICE_PATH     *DevicePath = 0;
UINT32                  UGAOnly = 0;
UINT32                 SafeMode = 0;
UINT32                   Mem640 = 0;       // 1Mb actual memory size
struct gate64_s      *SavedIDTv = 0;
void                 *XcptStack = 0;
extern u64t        trap_table[];

static UINT8      int_used[256];           // "vector was changed" flag

#if PRINT_MEM_TABLE > 0
static CHAR16  *MemTypeStr[EfiMaxMemoryType]  = {
            L"reserved  ",
            L"LoaderCode",
            L"LoaderData",
            L"BS_code   ",
            L"BS_data   ",
            L"RT_code   ",
            L"RT_data   ",
            L"available ",
            L"Unusable  ",
            L"ACPI_recl ",
            L"ACPI_NVS  ",
            L"MemMapIO  ",
			L"MemPortIO ",
			L"PAL_code  "
    };
#endif

extern void timer_irq(void);
extern void iretq_addr(void);

EFI_STATUS OpenProtocolByHandle(EFI_HANDLE handle, EFI_GUID *protocol,
                                void **interface)
{
   return (BS->OpenProtocol(handle, protocol, interface, SelfImg, 0,
           EFI_OPEN_PROTOCOL_BY_HANDLE_PROTOCOL));
}

void ErrorExit(CHAR16 *msg) {
   mfs_term();
   vio_term();
   // here we get back EFI console
   Print(msg);
   dsk_term();
   // restore GDT
   if (GDTCopy) {
      GDT64  cGDT;
      ReadGDTR(&cGDT);
      Print (L"GDT base: %X, size %X\n", cGDT.base, cGDT.limit+1);
      // if GDT is ours - restore original
      if (cGDT.base==GDTCopy) {
         RestSegRegs();
         WriteGDTR(&SavedGDT);
         Print (L"GDT restored\n");
      }
      // GDT buffer in EFI memory? Free it
      if (GDTCopyPages) BS->FreePages(GDTCopy, GDTCopyPages);
   }
   // free QSINIT memory space if allocated
   if (QSMem) BS->FreePages(QSMem, QSPages);
   // return to EFI shell/boot sequence
   BS->Exit(SelfImg, EFI_SUCCESS, 0, NULL);
}

void init_memory(void) {
   EFI_STATUS             Status;
   EFI_MEMORY_DESCRIPTOR    *MME,
                      *MemoryMap;
   UINTN               NoEntries;
   UINTN                  MapKey;
   UINTN                DescSize;
   UINT32                DescVer;
   EFI_PHYSICAL_ADDRESS    RcMem = 0;
   UINTN                 RcPages = 0, 
                       MaxContig = 0;
   UINTN               TotalFree = 0;
   UINTN                      ii;

   MemoryMap = LibMemoryMap(&NoEntries, &MapKey, &DescSize, &DescVer);
   if (!MemoryMap) ErrorExit(L"Can not retrieve the current memory map\n");

#if PRINT_MEM_TABLE > 0
   Print(L"Memory Descriptor List:\n\n");
   Print(L"  Type        Start Address     End Address       Attributes      \n");
   Print(L"  ==========  ================  ================  ================\n");
#endif
   MME = MemoryMap;
   /* calculate:
      - number of free pages below 4Gb 
      - max contigous block size
      - actual size of 640k in 1st Mb */
   for (ii=0; ii<NoEntries; ii++) {
      if (MME->PhysicalStart<_1MB && (MME->Type==EfiConventionalMemory ||
         MME->Type==EfiLoaderCode || MME->Type==EfiLoaderData ||
           MME->Type==EfiBootServicesCode || MME->Type==EfiBootServicesData ||
              MME->Type==EfiRuntimeServicesCode || MME->Type==EfiRuntimeServicesData))
      {
         EFI_PHYSICAL_ADDRESS bend = MME->PhysicalStart+(MME->NumberOfPages<<EFI_PAGE_SHIFT);
         if (bend<=0xA0000 && bend>Mem640) Mem640 = (UINT32)bend;
      }
      if (MME->Type==EfiConventionalMemory && MME->PhysicalStart>=_1MB &&
         MME->PhysicalStart<_4GBLL) 
      {
         TotalFree+=MME->NumberOfPages;
         if (MME->NumberOfPages>MaxContig) MaxContig = MME->NumberOfPages;
      }
#if PRINT_MEM_TABLE > 0
      Print(L"  %s  %X  %X  %X\n", MemTypeStr[MME->Type], MME->PhysicalStart,
            MME->PhysicalStart+(MME->NumberOfPages<<EFI_PAGE_SHIFT)-1, MME->Attribute);
#endif
      MME = NextMemoryDescriptor(MME, DescSize);
   }
   // try to allocate block at the top of 4Gb!
   MME = MemoryMap;
   for (ii=0; ii<NoEntries; ii++) {
      if (MME->Type==EfiConventionalMemory && MME->NumberOfPages<<EFI_PAGE_SHIFT>=_32MB &&
         MME->PhysicalStart>=_64MB && MME->PhysicalStart<_4GBLL)
      {
         EFI_ALLOCATE_TYPE target = AllocateAddress;
         RcPages = MME->NumberOfPages;
         RcMem   = MME->PhysicalStart;

         /* if block >1.5Gb or >1/2 of all available memory - use 1/2 of it */
         if (RcPages>_1GB+_512MB>>PAGESHIFT || RcPages>TotalFree/2 ) {
            RcPages = RcPages/2;
            target  = AllocateMaxAddress;
            RcMem   = FFFF & ~PAGEMASK;
         }
         Status = BS->AllocatePages(target, EfiLoaderData, RcPages, &RcMem);
         if (EFI_ERROR(Status)) RcMem = 0; else break;
      }
      MME = NextMemoryDescriptor(MME, DescSize);
   }
   /* failed to allocate at lowest address? then try to use:
      1. 2/3 of max contigous below 4Gb
      2. 32Mb below 4Gb
      3. 32Mb anywhere */
   if (!RcMem) {
      for (ii=0; ii<3; ii++) {
         RcMem   = FFFF & ~PAGEMASK;
         RcPages = ii ? _32MB>>EFI_PAGE_SHIFT : MaxContig*2/3;

         Status  = BS->AllocatePages(ii<2?AllocateMaxAddress:AllocateAnyPages,
            EfiLoaderData, RcPages, &RcMem);
         if (!EFI_ERROR(Status)) break; else RcMem = 0;
      }
   }
   FreePool(MemoryMap);
   // check result
   if (!RcMem || RcPages < _32MB>>EFI_PAGE_SHIFT) ErrorExit(L"No 32Mb of free memory!\n");
   if (RcMem>=_4GBLL) ErrorExit(L"Unable to allocate memory below 4Gb!\n");
   // finally, we got it!
   if (RcMem && RcPages) {
      QSMem   = RcMem;
      QSPages = RcPages;
   }
}


EFI_PHYSICAL_ADDRESS PageAlloc(u32t pages) {
   EFI_PHYSICAL_ADDRESS Addr = FFFF & ~PAGEMASK;
   EFI_STATUS         Status = BS->AllocatePages(AllocateMaxAddress, EfiLoaderData, 
                                                 pages, &Addr);
   if (EFI_ERROR(Status)) ErrorExit(L"Memory allocation error!\n");
   return Addr;
}

u32t XcptOn(void) {
   if (!SafeMode) {
      struct gate64_s *idesc = (struct gate64_s *)CurIDT.base;
      u16t cs64 = GetCodeSelector();
      int    ii = 0;

      IntsOff();
      while (trap_table[ii]!=FFFF64) {
         u64t inta = trap_table[ii];
         if (inta) {
            idesc->g64_sel    = cs64;
            idesc->g64_ofs0   = (u16t)inta;
            idesc->g64_ofs16  = (u16t)(inta>>16);
            idesc->g64_ofs32  = (u32t)(inta>>32);
            idesc->g64_parms  = 0;
            idesc->g64_access = D_PRES|D_TRAPGATE32;
            // mark it to restore
            int_used[ii] = 1;
         }
         ii++;
         idesc++;
      }
      IntsOn();
      return 1;
   }
   return 0;
}

u32t TimerIrqsOn(u32t* apic, u8t tmr_vector, u8t spr_vector) {
   if (tmr_vector<16 || spr_vector<16) return 0; else {
      u16t cs64 = GetCodeSelector();
      int    ii;

      IntsOff();
      lapic    = apic;
      SavedSpr = apic[APIC_SPURIOUS];

      for (ii=0; ii<2; ii++) {
         u32t   vn = ii?spr_vector:tmr_vector;
         u64t inta = (u64t)(ii?&iretq_addr:&timer_irq);
         struct gate64_s *idesc = (struct gate64_s *)CurIDT.base + vn;
         idesc->g64_sel    = cs64;
         idesc->g64_ofs0   = (u16t)inta;
         idesc->g64_ofs16  = (u16t)(inta>>16);
         idesc->g64_ofs32  = (u32t)(inta>>32);
         idesc->g64_parms  = 0;
         idesc->g64_access = D_PRES|D_INTGATE32;
         // mark it to restore
         int_used[vn] = 1;
      }
      IntsOn();

      return 1;
   }
}

/* Function replaces EFI allocated block for GDT, IDT backup & stack to
   the buffer, provided by 32-bit part (in QSINIT memory area).

   Called immediately after ExitBootServices() to make sure, that GDT will
   stay valid when all EfiLoaderData memory will be marked as "free".

   Buffer is 128k, this should be enough (max 64k GDT + 4k IDT + stack) */
void ReplaceGDT(void *buffer) {
   if (GDTCopyPages) {
      EFI_PHYSICAL_ADDRESS newGDT = (EFI_PHYSICAL_ADDRESS)buffer;
      struct gate64_s *    newIDT = (struct gate64_s *)(newGDT + GDTSize);
      // disable ints to be safe
      IntsOff();
      RtCopyMem(buffer, (void*)GDTCopy, GDTSize);
      RtCopyMem(newIDT, SavedIDTv, sizeof(struct gate64_s)*256);

      Header32->gdt_pos = (u32t)newGDT;

      XcptStack    = (u8t*)newGDT + _128KB;
      SavedIDTv    = newIDT;
      GDTCopy      = newGDT;

      CurGDT.base  = GDTCopy;
      CurGDT.limit = GDTSize - 1;

      WriteGDTR(&CurGDT);

      IntsOn();
      // free memory is just not possible, because no boot services ...
      GDTCopyPages = 0;
   }
}

void RestoreIDT(void) {
   struct gate64_s *idesc = (struct gate64_s *)CurIDT.base;
   int  ii;

   IntsOff();
   for (ii=0; ii<256; ii++)
      if (int_used[ii]) idesc[ii] = SavedIDTv[ii];
   IntsOn();
}

// "main" function
void QSLoader(IN EFI_HANDLE ImgHandle, IN EFI_SYSTEM_TABLE *SysTab) {
   EFI_STATUS   st;
   UINTN        ii;
   // Initialize the Library. Set BS, RT, &ST globals
   //  BS = Boot Services RT = RunTime Services
   //  ST = System Table
   InitializeLib (ImgHandle, SysTab);

   SelfImg = ImgHandle;
   // save and replace GDT 
   ReadGDTR(&SavedGDT);
   
   if (0xFFFF - SavedGDT.limit < RESERVE_DESCRIPTORS * 8)
      ErrorExit(L"There is no space in System GDT, unable to run!\n");
   // who can allocate above 4Gb? still care
   GDTSize      = SavedGDT.limit + 1 + RESERVE_DESCRIPTORS * 8;
   if (GDTSize>_64KB) GDTSize = _64KB;
   GDTCopyPages = Round4k(GDTSize + sizeof(struct gate64_s)*256 + XCPT_STACK_SIZE)>>12;
   GDTCopy      = PageAlloc(GDTCopyPages);
   // makes a copy of original GDT
   RtZeroMem((void*)GDTCopy, GDTSize);
   RtCopyMem((void*)GDTCopy, (void*)SavedGDT.base, GDTLowest = SavedGDT.limit + 1);

   RtZeroMem(&int_used, sizeof(int_used));

   SavedIDTv    = (struct gate64_s *)(GDTCopy + GDTSize);
   // points to the end of block
   XcptStack    = (u8t*)(SavedIDTv + 256) + XCPT_STACK_SIZE;

   CurGDT.base  = GDTCopy;
   CurGDT.limit = GDTSize - 1;

   SaveSegRegs();
   WriteGDTR(&CurGDT);
#if PRINT_MEM_TABLE > 0
   Print(L"GDT base: %X, size %X\n", GDTCopy, GDTSize);
#endif
   BS->SetWatchdogTimer(0, 0, 0, 0);

   // Get the device handle and file path to the EFI OS Loader itself.
   st = BS->HandleProtocol(ImgHandle, &LoadedImageProtocol, (VOID*)&LoadedImage);
   if (EFI_ERROR(st))
      ErrorExit(L"Can not retrieve a LoadedImageProtocol handle for ImageHandle\n");

   st = BS->HandleProtocol(LoadedImage->DeviceHandle, &DevicePathProtocol, (VOID*)&DevicePath);
   if (EFI_ERROR(st) || DevicePath==NULL)
      ErrorExit(L"Can not find a DevicePath handle for LoadedImage\n");

   // primitive command line parser
   if (LoadedImage->LoadOptionsSize) {
	  CHAR16 *cmdline = LoadedImage->LoadOptions,
	             nopt[32];
	  UINT32  ii, optc = 0;

	  for (ii=0; ii<LoadedImage->LoadOptionsSize; ii++)
	     if (cmdline[ii]<=' ') {
	        if (optc) {
	           nopt[optc] = 0;
	           if (RtStrICmp(nopt,L"/UGA")==0) UGAOnly = 1; else
	           if (RtStrICmp(nopt,L"/SM")==0) SafeMode = 1;
	           optc = 0;
	        }
	        if (!cmdline[ii]) break;
	     } else
	     if (optc<31) nopt[optc++] = cmdline[ii];
   }
   // init keyboard
   key_init();
   // init console
   vio_init();
   // calculate "timer" ticks
   tm_counter();
   // find a place to place 32-bit part ;)
   init_memory();
#if PRINT_MEM_TABLE > 0
   Print(L"  QSINIT memory at address                   : %X\n", QSMem);
   Print(L"         size in kb                          : %ld\n", ((u64t)QSPages<<PAGESHIFT)>>10);

   Print(L"\nPress ENTER to continue...\n");
   {
      u32t key;
      do {
         key = key_read();
         Print (L"Key Code : %X\n", key);
      } while ((key&0xFF) != 13);
   } 
#endif
   if (!SafeMode) {
      // check for key at least once and hope it will return LEFT SHIFT for us ;)
      key_pressed();
      if (key_status()&KEY_SHIFTLEFT) SafeMode = 1;
   }
   // save IDT
   ReadIDTR(&CurIDT);
   RtCopyMem(SavedIDTv, (void*)CurIDT.base, sizeof(struct gate64_s)*256);
   // save CR4
   SavedCr4 = hlp_setreg(4, FFFF64);

   ii = Load32();
   if (ii) Print(L"Load32() rc = %d\n",ii);
   ii = SysConfig();
   if (ii) Print(L"SysConfig() rc = %d\n",ii);
   ii = VioConfig();
   if (ii) Print(L"VioConfig() rc = %d\n",ii);
   ii = Exec32();
   // restore APIC
   if (lapic) {
      IntsOff();
      lapic[APIC_LVT_TMR]  = APIC_DISABLE;
      lapic[APIC_SPURIOUS] = SavedSpr;
      IntsOn();
   }
   if (ii) Print(L"Exec32() rc = %d\n",ii);

   RestoreIDT();
   // restore cr4
   hlp_setreg(4, SavedCr4);

   ErrorExit(L"Bye!\n");
}
